// ----------------------------
// Variables and User Variables 
// ---------------------------- 

.pc = $0002 "zeropage first usable address" virtual 

// zero page variable definitions and library initialization
// ("fast RAM" variables)    
// 
.if ( include_covert_bitops_loader == 1 )
{
  .pc = $0002 "zeropage covert bitops loadersytem vars" virtual
  // loader needs at least 8..
  .fill 8, 0
  .if ( ADDITIONAL_ZEROPAGE == 1 )
  {
    // ..another 4 in some configs
    .fill 4, 0
  }
}
.pc = * "zeropage cavelib vars" virtual
cl_w0: .word $FFFF // general purpose 16 bit value 0
cl_w1: .word $FFFF // general purpose 16 bit value 1
cl_w2: .word $FFFF // general purpose 16 bit value 2
cl_w3: .word $FFFF // general purpose 16 bit value 3
cl_b0: .byte $FF   // general purpose  8 bit value 0
cl_b1: .byte $FF   // general purpose  8 bit value 1
cl_b2: .byte $FF   // general purpose  8 bit value 2
cl_b3: .byte $FF   // general purpose  8 bit value 3
cl_b4: .byte $FF   // general purpose  8 bit value 4
// TODO: add more cavelib variables here

.label cavelib_vars_end = *

.var cv_clear_pattern = $00
.var cavelib_user_vars_end = $00

.var cavelib_routines_ready = 0

// set all cavelib vars to the clear pattern
.pseudocommand init_cavelib_vars 
{
  :zp_memset #$02 ; #cavelib_vars_end-2 ; #cv_clear_pattern
}


// set all cavelib user vars to the clear pattern
.pseudocommand init_cavelib_user_vars
{
  .if (cavelib_user_vars_end == $00 || cavelib_user_vars_begin == $00 )
  {
    .print "zeropage variable definitions example:"
    .print ":begin_zp_user_vars"
    .print "my_array: .fill $20, 0"
    .print ":end_zp_user_vars"
    .error "cavelib user must enclose zero page variables as shown above" 
  }
   
  .if ( cavelib_user_vars_end-1 > $ff )
  {
    .printnow "first zeropage user var: " + address_of(cavelib_user_vars_begin)
    .printnow "   address behind those: " + address_of(cavelib_user_vars_end)
  
    .error "zp user vars may not exceed $" + toHexString(cavelib_vars_end,2) + "-$ff range" 
  }
  
  :zp_memset #cavelib_vars_end ; #cavelib_user_vars_end-cavelib_vars_end; #cv_clear_pattern
}


// the following pseudocommand must be used to set the .pc in code  
// so that variables will then be located after the cavelib
// variables inside zero page RAM
// NOTE: by design, none of these vars can be initialized at compile-time
//       because they are located in the zero page locations used by BASIC
//       so all zero page variables have to be initialized manually at
//       runtime after BASIC ROM has been disabled (e.g. by :init_cavelib
.var cavelib_user_vars_begin = $00
.pseudocommand begin_zp_user_vars 
{ 
  .pc = cavelib_vars_end "zeropage user vars" virtual 
  .eval cavelib_user_vars_begin = *
}


// this pseudocommand must be placed after the user zero page variables
.pseudocommand end_zp_user_vars
{
  .eval cavelib_user_vars_end = *
  
  // generate some memory map info for covert bitops loader system here
  .if ( include_covert_bitops_loader == 1 )
  {
    .pc = loadbuffer "covert bitops loadersystem loadbuffer" virtual
    .fill $100, 0
    
    .if ( LOADFILE_PUCRUNCH > 0 )
    {
      .pc = depackbuffer "covert bitops loadersystem pucrunch depackbuffer" virtual
      .fill $1F, 0
    }
    .if ( LOADFILE_EXOMIZER > 0 )
    {
      .pc = depackbuffer "covert bitops loadersystem exomizer depackbuffer" virtual
      .fill $9C, 0
    }
  }
}


// start defining non zero page variables and routines after using this command 
// and they will be placed shortly behind the Basic Upstart 
.pseudocommand begin_slow_ram_vars_and_routines
{
 .if( cavelib_routines_ready == 0)
 {
  .eval cavelib_routines_ready = 1

  .pc = $0810 "slow ram cavelib - vars and routines"
  .eval cavelib_basic_zp_saved = * 
  .pc = * "slow ram cavelib - basic working area store/restore" 
  .fill $8e , 0 // for storing/restoring BASIC working area 02-8f from zeropage
  
  .eval cavelib_kernal_zp_saved = * 
  .pc = * "slow ram cavelib - kernal working area store/restore"
  .fill $70 , 0 // for storing/restoring KERNAL working area 90-ff from zeropage 
  
  // place mem subroutines
  .pc = * "slow ram cavelib - memory subroutines"
  :place_subroutine_memset
  :place_subroutine_memxor
  :place_subroutine_memcpy
  :place_subroutine_gapcpy
  :place_subroutine_memswap
  
  // place irq subroutines
  .pc = * "slow ram cavelib - nmi/irq/brk/end_irq handlers"
  :place_nmi_handler
  :place_irq_brk_handler
  :place_end_irq_handler
  
  // place misc helper subroutines 
  .pc = * "slow ram cavelib - misc helper subroutines"
  :place_subroutine_cstring_length
  :place_subroutine_waste_cycles_all

  // place file subroutines
  .pc = * "slow ram cavelib - file subroutines"
  :place_subroutine_load_error_store
  :place_subroutine_load_sync
  
  // place gfx subroutines
  .pc = * "slow ram cavelib - gfx subroutines"
  :place_subroutine_copy_char_rom
  :place_subroutine_textout1x2
  :place_subroutine_textout1x2_to_sprites

  // TODO: place more cavelib specific variables and routines here
  
  // place 3-rd party subroutines
  .if ( include_covert_bitops_loader == 1 )
  {
    .pc = * "slow ram cavelib - Covert Bitops Loadersystem V2.26ka"
    :place_covert_bitops_loader_code
  }
  
  .pc = * "slow ram user - vars and routines"
  
 }
}


// place this behind non zero page variables and routines definitions and before the 
// main program code to have the memory map show the address range
// used by the program
.pseudocommand end_slow_ram_vars_and_routines
{
  .pc = * "main program"
}
